package dev.shadowsoffire.apotheosis.adventure.boss;

import java.util.List;
import dev.shadowsoffire.placebo.json.NBTAdapter;
import net.minecraft.nbt.Tag;
import net.minecraft.nbt.NbtUtils;
import com.mojang.datafixers.kinds.Applicative;
import java.util.function.Function;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import java.util.Set;
import dev.shadowsoffire.apotheosis.Apotheosis;
import com.mojang.serialization.Codec;
import javax.annotation.Nullable;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.world.entity.MobSpawnType;
import net.minecraft.world.level.ServerLevelAccessor;
import net.minecraft.world.entity.Mob;
import dev.shadowsoffire.placebo.codec.CodecMap;
import dev.shadowsoffire.placebo.codec.CodecProvider;

public interface Exclusion extends CodecProvider<Exclusion>
{
    public static final CodecMap<Exclusion> CODEC = new CodecMap("Miniboss Exclusion");
    
    boolean isExcluded(final Mob p0, final ServerLevelAccessor p1, final MobSpawnType p2, @Nullable final CompoundTag p3);
    
    boolean requiresNbtAccess();
    
    default void initSerializers() {
        register("spawn_type", SpawnTypeExclusion.CODEC);
        register("nbt", NbtExclusion.CODEC);
        register("surface_type", SurfaceTypeExclusion.CODEC);
        register("and", AndExclusion.CODEC);
    }
    
    default void register(final String id, final Codec<? extends Exclusion> codec) {
        Exclusion.CODEC.register(Apotheosis.loc(id), (Codec)codec);
    }
    
    record SpawnTypeExclusion(Set<MobSpawnType> types) implements Exclusion {
        public static Codec<SpawnTypeExclusion> CODEC;
        
        public Codec<? extends Exclusion> getCodec() {
            return SpawnTypeExclusion.CODEC;
        }
        
        @Override
        public boolean isExcluded(final Mob mob, final ServerLevelAccessor level, final MobSpawnType spawnType, final CompoundTag entityNbt) {
            return this.types.contains(spawnType);
        }
        
        @Override
        public boolean requiresNbtAccess() {
            return false;
        }
        
        static {
            SpawnTypeExclusion.CODEC = (Codec<SpawnTypeExclusion>)RecordCodecBuilder.create(inst -> inst.group((App)PlaceboCodecs.setOf(PlaceboCodecs.enumCodec((Class)MobSpawnType.class)).fieldOf("spawn_types").forGetter((Function)SpawnTypeExclusion::types)).apply((Applicative)inst, (Function)SpawnTypeExclusion::new));
        }
    }
    
    record NbtExclusion(CompoundTag nbt) implements Exclusion {
        public static Codec<NbtExclusion> CODEC;
        
        public Codec<? extends Exclusion> getCodec() {
            return NbtExclusion.CODEC;
        }
        
        @Override
        public boolean isExcluded(final Mob mob, final ServerLevelAccessor level, final MobSpawnType spawnType, final CompoundTag entityNbt) {
            return NbtUtils.m_129235_((Tag)this.nbt, (Tag)entityNbt, true);
        }
        
        @Override
        public boolean requiresNbtAccess() {
            return true;
        }
        
        static {
            NbtExclusion.CODEC = (Codec<NbtExclusion>)RecordCodecBuilder.create(inst -> inst.group((App)NBTAdapter.EITHER_CODEC.fieldOf("nbt").forGetter((Function)NbtExclusion::nbt)).apply((Applicative)inst, (Function)NbtExclusion::new));
        }
    }
    
    record SurfaceTypeExclusion(BossEvents.BossSpawnRules rule) implements Exclusion {
        public static Codec<SurfaceTypeExclusion> CODEC;
        
        public Codec<? extends Exclusion> getCodec() {
            return SurfaceTypeExclusion.CODEC;
        }
        
        @Override
        public boolean isExcluded(final Mob mob, final ServerLevelAccessor level, final MobSpawnType spawnType, final CompoundTag entityNbt) {
            return !this.rule.test(level, mob.m_20183_());
        }
        
        @Override
        public boolean requiresNbtAccess() {
            return false;
        }
        
        static {
            SurfaceTypeExclusion.CODEC = (Codec<SurfaceTypeExclusion>)RecordCodecBuilder.create(inst -> inst.group((App)BossEvents.BossSpawnRules.CODEC.fieldOf("rule").forGetter((Function)SurfaceTypeExclusion::rule)).apply((Applicative)inst, (Function)SurfaceTypeExclusion::new));
        }
    }
    
    record AndExclusion(List<Exclusion> exclusions) implements Exclusion {
        public static Codec<AndExclusion> CODEC;
        
        public Codec<? extends Exclusion> getCodec() {
            return AndExclusion.CODEC;
        }
        
        @Override
        public boolean isExcluded(final Mob mob, final ServerLevelAccessor level, final MobSpawnType spawnType, final CompoundTag entityNbt) {
            return this.exclusions.stream().allMatch(e -> e.isExcluded(mob, level, spawnType, entityNbt));
        }
        
        @Override
        public boolean requiresNbtAccess() {
            return this.exclusions.stream().anyMatch(Exclusion::requiresNbtAccess);
        }
        
        static {
            AndExclusion.CODEC = (Codec<AndExclusion>)RecordCodecBuilder.create(inst -> inst.group((App)Exclusion.CODEC.listOf().fieldOf("exclusions").forGetter((Function)AndExclusion::exclusions)).apply((Applicative)inst, (Function)AndExclusion::new));
        }
    }
}
